from ast import literal_eval
from configparser import ConfigParser
from warnings import warn
import json
import yaml


def parse_ini(content):
    config = ConfigParser(allow_no_value=True)
    config.read_string(content)
    if len(config.sections()) == 0:
        return dict(config.items("DEFAULT"))
    result = {}
    for section in config.sections():
        result[section] = dict(config.items(section))
    return result


class File:
    ENCODINGS = {
        "ini": parse_ini,
        "json": json.loads,
        "raw": literal_eval,
        "yaml": yaml.safe_load,
    }

    def __init__(self, path, encoding="raw", parser=None):
        self.__read_file(path)
        self.__set_encoding(encoding, parser)
        self.__parse_content()
        self.__clear_empty_values()

    def get(self):
        return self.content

    def __read_file(self, path):
        try:
            with open(path, "r") as f:
                self.content = f.read()
        except IOError:
            warn("IOError when opening {}".format(path), UserWarning)
            self.content = {}

    def __set_encoding(self, encoding, parser=None):
        try:
            self.parser = self.ENCODINGS[encoding]
        except KeyError:
            self.parser = parser

    def __parse_content(self):
        try:
            self.content = self.parser(self.content)
            if type(self.content) != dict:
                self.content = {}
        except Exception:
            self.content = {}

    def __clear_empty_values(self):
        if not self.content:
            return
        keys_to_clear = []
        for key, value in self.content.items():
            if value is None:
                keys_to_clear.append(key)
        for key in keys_to_clear:
            self.content.pop(key, None)



import os
import re
from ast import literal_eval
from warnings import warn


class Env(object):
    def __init__(
        self,
        separator=None,
        match=None,
        whitelist=None,
        parse_values=False,
        to_lower=False,
        convert_underscores=False,
        docker_secrets=None,
    ):
        self.separator = separator
        self.match = match
        self.whitelist = whitelist
        self.parse_values = parse_values
        self.to_lower = to_lower
        self.convert_underscores = convert_underscores
        self.docker_secrets = docker_secrets

        if self.match is not None:
            self.re = re.compile(self.match)

        self.__gather_vars()

        if self.to_lower:
            self.vars = self.__change_keys(self.vars, self.__to_lower)

        if self.convert_underscores:
            self.vars = self.__change_keys(self.vars, self.__convert_underscores)

    def get(self):
        return self.vars

    def __valid_key(self, key):
        if self.match is not None and self.whitelist is not None:
            return key in self.whitelist or self.re.search(key) is not None
        elif self.match is not None:
            return self.re.search(key) is not None
        elif self.whitelist is not None:
            return key in self.whitelist
        else:
            return True

    def __split_vars(self, env_vars):
        keys_to_delete = []
        dict_to_add = {}
        for key in env_vars.keys():
            splits = key.split(self.separator)
            splits = list(filter(None, splits))

            if len(splits) != 1:
                split = self.__split_var(splits, env_vars[key])
                keys_to_delete.append(key)
                self.__merge_split(split, dict_to_add)

        for key in keys_to_delete:
            del env_vars[key]
        env_vars.update(dict_to_add)

    def __split_var(self, keys, value):
        if len(keys) == 1:
            return {keys[0]: value}
        else:
            key = keys[0]
            del keys[0]
            return {key: self.__split_var(keys, value)}

    def __merge_split(self, split, env_vars):
        key = list(split.keys())[0]
        value = list(split.values())[0]
        if key not in env_vars:
            env_vars[key] = value
            return
        elif type(value) == dict:
            self.__merge_split(value, env_vars[key])
        else:
            return

    def __try_parse(self, env_vars):
        for key, value in env_vars.items():
            try:
                if value.lower() == "true":
                    env_vars[key] = True
                elif value.lower() == "false":
                    env_vars[key] = False
                else:
                    env_vars[key] = literal_eval(value)
            except (ValueError, SyntaxError):
                pass

    def __handle_docker_secret(self, key, value):
        postfix = "_FILE"
        if key.endswith(postfix):
            try:
                with open(value, "r") as f:
                    self.vars[key[0 : -len(postfix)]] = f.read().strip()  # noqa: E203
            except IOError:
                warn("IOError when opening {}".format(value), UserWarning)

    def __gather_vars(self):
        self.vars = {}
        env_vars = os.environ

        for key in env_vars.keys():
            if self.__valid_key(key):
                if self.docker_secrets is not None and key in self.docker_secrets:
                    self.__handle_docker_secret(key, env_vars[key])
                else:
                    self.vars[key] = env_vars[key]

        if self.parse_values:
            self.__try_parse(self.vars)

        if self.separator is not None:
            self.__split_vars(self.vars)

    def __to_lower(self, key):
        return key.lower()

    def __convert_underscores(self, key):
        return key.replace("_", "-")

    def __change_keys(self, env_vars, operation):
        new_dict = {}
        for key, value in env_vars.items():
            if type(value) == dict:
                new_dict[operation(key)] = self.__change_keys(value, operation)
            else:
                new_dict[operation(key)] = env_vars[key]
        return new_dict



import argparse
from ast import literal_eval


class Argv(object):
    parser = None

    def __init__(self, name, short_name=None, type=None, help=None):
        if not name.startswith("-") or name.count(" ") != 0:
            raise ValueError

        if Argv.parser is None:
            Argv.parser = argparse.ArgumentParser()

        self.results = {}

        args = {}
        args["help"] = help
        if type == bool:
            args["action"] = "store_true"
        # types supported by literal_eval
        elif type in [dict, list, tuple]:
            args["type"] = literal_eval
        elif type == "repeated_list":
            args["action"] = "append"
        else:
            args["type"] = type

        if name.lstrip("-").count("-") != 0:
            args["dest"] = name.lstrip("-")

        if short_name is not None:
            Argv.parser.add_argument(
                name, short_name, default=argparse.SUPPRESS, **args
            )
        else:
            Argv.parser.add_argument(name, default=argparse.SUPPRESS, **args)

    def get(self):
        namespace, _ = Argv.parser.parse_known_args()
        self.results = vars(namespace)
        return self.results



from .store import argv
from .store import env
from .store import file
from .store import memory

from deepmerge import Merger


class Pconf(object):
    """Entry point class for the pconf module.

    Doesn't need to be instantiated. First set the hierarchy, by calling the
    desired source methods in the desired order. The source set first has the
    highest priority.

    After sources are set the values can be accessed by calling get()
    """

    __hierarchy = []

    merger = Merger([(dict, ["merge"])], ["override"], ["override"])

    @classmethod
    def get(cls):
        """Get values gathered from the previously set hierarchy.

        Respects the order in which sources are set, the first source set
        has the highest priority, overrides values with the same key that
        exist in sources with lower priority.

        Returns:
            dict: The dictionary containing values gathered from all set sources.
        """
        results = {}

        hierarchy = cls.__hierarchy[::-1]

        for storeMethod in hierarchy:
            cls.merger.merge(results, storeMethod.get())

        return results

    @classmethod
    def defaults(cls, data):
        """Set passed values as a source

        Stores the passed dict in memory, use to manually set default values
        in the hierarchy, by setting this in the end.

        Args:
            data: the dict to store
        """
        cls.__hierarchy.append(memory.Memory(data))

    @classmethod
    def overrides(cls, data):
        """Set passed values as a source

        Stores the passed dict in memory, use to manually override any values
        in the hierarchy, by setting this in the front.

        Args:
            data: the dict to store
        """
        cls.__hierarchy.append(memory.Memory(data))

    @classmethod
    def argv(cls, name, short_name=None, type=None, help=None):
        """Set command line arguments as a source

        Parses the command line arguments described by the parameters.

        Args:
            name: the long name of the argument (foo)
            short_name: the optional short name of the argument (f)
            type: the optional type of the argument, defaults to bool
            help: the optional help text for the argument
        """
        cls.__hierarchy.append(argv.Argv(name, short_name, type, help))

    @classmethod
    def env(
        cls,
        separator=None,
        match=None,
        whitelist=None,
        parse_values=None,
        to_lower=None,
        convert_underscores=None,
        docker_secrets=None,
    ):
        """Set environment variables as a source.

        By default all environment variables available to the process are used.
        This can be narrowed by the args.

        Args:
            separator: Keys are split along this character, the resulting
                splits are considered nested values.
            match: Regular expression for key matching. Keys matching the
                expression are considered whitelisted.
            whitelist: Only use environment variables that are listed in this
                list.
            parse_values: Try to parse all variable for well-known types.
            to_lower: Convert all variable names to lower case.
            convert_underscores: Convert all underscores in the name to dashes,
                this takes place after separation via the separator option.
            docker_secrets: A list of names with _FILE postfix, which will have
                their postfix removed and the content of the file pointed by
                their original value.
        """
        cls.__hierarchy.append(
            env.Env(
                separator,
                match,
                whitelist,
                parse_values,
                to_lower,
                convert_underscores,
                docker_secrets,
            )
        )

    @classmethod
    def file(cls, path, encoding=None, parser=None):
        """Set a file as a source.

        File are parsed as literal python dicts by default, this behaviour
        can be configured.

        Args:
            path: The path to the file to be parsed
            encoding: The encoding of the file.
                Defaults to 'raw'. Available built-in values: 'ini', 'json', 'yaml'.
                Custom value can be used in conjunction with parser.
            parser: A parser function for a custom encoder.
                It is expected to return a dict containing the parsed values
                when called with the contents of the file as an argument.
        """
        cls.__hierarchy.append(file.File(path, encoding, parser))

    @classmethod
    def clear(cls):
        """Purge everything from the current hierarchy.

        In case of the need for multiple configurations this can be used to
        reset everything to a blank slate and start over.
        """
        del cls.__hierarchy[:]  # instead of clear() to maintain 2.7 compatibility



from os import path
from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(path.join(path.dirname(__file__), fname)).read()


setup(
    name="pconf",
    version="1.10.0",
    author="Andras Maroy",
    author_email="andras@maroy.hu",
    description=("Hierarchical python configuration with files, environment variables, command-line arguments."),
    license="MIT",
    keywords="configuration hierarchical",
    url="https://github.com/andrasmaroy/pconf",
    packages=['pconf', 'pconf.store'],
    long_description=read('README.rst'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11'
    ],
    install_requires=['pyyaml', 'deepmerge'],
    extras_require={
        'test': ['pytest', 'mock'],
    },
)
